#!/usr/bin/perl
#/gsc/software/linux-x86_64-centos5/perl-5.16.3/bin/perl

#AUTHOR
#   Rene Warren
#   rwarren at bcgsc.ca

#NAME
#LINKS: Long Interval Nucleotide K-mer Scaffolder 

#SYNOPSIS

#DOCUMENTATION
#   LINKS-readme.txt distributed with this software @ www.bcgsc.ca
#   http://www.bcgsc.ca/platform/bioinfo/software/links
#   We hope this code is useful to you -- Please send comments & suggestions to rwarren * bcgsc.ca
#   If you use LINKS, the LINKS code or ideas, please cite our work

#LICENSE
#   LINKS Copyright (c) 2014-2015 Canada's Michael Smith Genome Science Centre.  All rights reserved.
#   LINKS is released under the GNU General Public License


use strict;
use POSIX;
use FindBin;
#perl 5.8.8
#use lib "$FindBin::Bin/./lib/Bloom-Faster-1.6/bloom/lib64/perl5/site_perl/5.8.8/x86_64-linux-thread-multi";
#perl 5.16.3
use lib "$FindBin::Bin/./lib/Bloom-Faster-1.7/bloom5-16-3/lib/site_perl/5.16.3/x86_64-linux";###rebuild against PERL version or replace by pre-built
use Bloom::Faster;
use Getopt::Std;
use Net::SMTP;
use vars qw($opt_f $opt_s $opt_d $opt_k $opt_e $opt_l $opt_a $opt_v $opt_b $opt_t $opt_p $opt_o $opt_r $opt_x);
getopts('f:s:d:k:e:l:a:v:b:t:p:o:r:x');
my ($bf_file,$base_name,$frag_dist,$k,$insert_stdev,$min_links,$max_link_ratio,$verbose,$step,$offset,$fpr,$bfoff)=("","",4000,15,0.1,5,0.3,0,2,0,0.0001,0);

my $version = "[v1.5]";
my $dev = "rwarren\@bcgsc.ca";

#-------------------------------------------------

if(! $opt_f || ! $opt_s){
   print "Usage: $0 $version\n";
   print "-f  sequences to scaffold (Multi-FASTA format, required)\n"; 
   print "-s  file-of-filenames, full path to long sequence reads (Multi-FASTA/fastq format, required)\n";
   print "-d  distance between k-mer pairs (eg. target distance to re-scaffold on. default -d $frag_dist, optional)\n";
   print "-k  k-mer value (default -k $k, optional)\n";
   print "-t  step of sliding window when extracting k-mer pairs from long reads (default -t $step, optional)\n"; 
   print "-o  offset position for extracting k-mer pairs (default -o $offset, optional)\n";
   print "-e  error (%) allowed on -d distance   e.g. -e 0.1  == distance +/- 10% (default -e $insert_stdev, optional)\n";
   print "-l  minimum number of links (k-mer pairs) to compute scaffold (default -l $min_links, optional)\n";
   print "-a  maximum link ratio between two best contig pairs (default -a $max_link_ratio, optional)\n";
   print "    *higher values lead to least accurate scaffolding*\n";
   print "-b  base name for your output files (optional)\n";
   print "-r  Bloom filter input file for sequences supplied in -s (optional, if none provided will output to .bloom)\n";
   print "    NOTE: BLOOM FILTER MUST BE DERIVED FROM THE SAME FILE SUPPLIED IN -f WITH SAME -k VALUE\n";
   print "    IF YOU DO NOT SUPPLY A BLOOM FILTER, ONE WILL BE CREATED (.bloom)\n";
   print "-p  Bloom filter false positive rate (default -p $fpr, optional; increase to prevent memory allocation errors)\n";
   print "-x  Turn off Bloom filter functionality (-x 1 = yes, default = no, optional)\n";
   die   "-v  Runs in verbose mode (-v 1 = yes, default = no, optional)\n"; 
}

my $assemblyfile = $opt_f;
my $longfile = $opt_s;
$frag_dist = $opt_d if($opt_d);
$k = $opt_k if($opt_k);
$verbose = $opt_v if($opt_v);
$min_links = $opt_l if($opt_l);
$max_link_ratio = $opt_a if($opt_a);
$step = $opt_t if($opt_t);
$insert_stdev = $opt_e if($opt_e);
$base_name = $opt_b if($opt_b);
$offset = $opt_o if($opt_o);
$bf_file = $opt_r if($opt_r);
$fpr = int($opt_p) if($opt_p);
$bfoff = $opt_x if($opt_x);

my $assemblyruninfo="";


if(! -e $assemblyfile){
   die "Invalid file: $assemblyfile -- fatal\n";
}


### Naming output files
if ($base_name eq ""){

   $base_name = $assemblyfile . ".scaff_s-" . $longfile . "_d" . $frag_dist . "_k" . $k . "_e" . $insert_stdev . "_l" . $min_links . "_a" . $max_link_ratio . "_t" . $step . "_o" . $offset . "_r-" . $bf_file . "_p" . $fpr . "_x" . $bfoff;

   my $pid_num = getpgrp(0);
   $base_name .= "_pid" . $pid_num;
}

my $log = $base_name . ".log";
my $scaffold = $base_name . ".scaffolds";
my $issues = $base_name . ".pairing_issues";
my $distribution = $base_name . ".pairing_distribution.csv";
my $bfout = $base_name . ".bloom";

open (LOG, ">$log") || die "Can't write to $log -- fatal\n";

#-------------------------------------------------
#
my $init_message = "\nRunning: $0 $version\n-f $assemblyfile\n-s $longfile\n";
$init_message .= "-d $frag_dist\n-k $k\n-e $insert_stdev\n-l $min_links\n-a $max_link_ratio\n-t $step\n-o $offset\n-b $base_name\n-r $bf_file\n-p $fpr\n-x $bfoff\n\n";

print $init_message;
print LOG $init_message;
$assemblyruninfo=$init_message . "\n\n";

#-------------------------------------------------
my $file_message = "";
my $ct_fof_line = 0;

if(! -e $longfile){
   $file_message = "Invalid file of filenames: $longfile -- fatal\n";
   print $file_message;
   print LOG $file_message;
   exit;
}else{
   open(FOF,$longfile) || die "Can't open file of filenames $longfile for reading -- fatal.\n";
   while(<FOF>){
      chomp;
      $ct_fof_line++;
      $file_message = "Checking $_...";
      print $file_message;
      print LOG $file_message;
      if(! -e $_){
         $file_message = "na\n*** File does not exist -- fatal (check the path/file and try again)\n";
         if(/^\>/){
           $file_message .= "It appears that the file you supplied in -s is in fasta format where it should be a file of filenames, listing the fullpath/fasta,fastq files to consider (1 per line).\n";
         }elsif(/^\@/){
           $file_message .= "It appears that the file you supplied in -s is in fastq format where it should be a file of filenames, listing the fullpath/fasta, fastq files to consider (1 per line).\n";
         }
         print $file_message;
         print LOG $file_message;
         exit;
      }else{
         $file_message = "ok\n";
         print $file_message;
         print LOG $file_message;
      }
   }
   close FOF;
}

if(! -e $assemblyfile){
   $file_message = "\nInvalid file: $assemblyfile -- fatal\n";
   print $file_message;
   print LOG $file_message;
   exit;
}else{
   $file_message = "Checking sequence target file $assemblyfile...ok\n";
   print $file_message;
   print LOG $file_message;
}

#-------------------------------------------------
#LINKS STARTS
#-------------------------------------------------

my $date = `date`;
chomp($date);
my $bloom;

eval{

my $reading_tigbloom_message = "\n=>Reading contigs file, building Bloom filter: $date\n";
print $reading_tigbloom_message;
print LOG $reading_tigbloom_message;
$assemblyruninfo.=$reading_tigbloom_message;
if($bfoff==0){
 if($bf_file eq ""){
   my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat($assemblyfile);
   my $bfelements = int($size);

   $date = `date`;
   chomp($date);

   $bfelements *= 2;###twice the kmer when looking at both strands
   my $m = ceil((-1 * $bfelements * log($fpr)) / (log(2) * log(2))); #the number of bits
   my $hashfct = floor(($m / $bfelements) * log(2));# the number of hash functions
   my $bfspecs = "*****\nBloom filter specs\nelements=$bfelements\nFPR=$fpr\nsize (bits)=$m\nhash functions=$hashfct\n*****\n";
   print $bfspecs;
   print LOG $bfspecs;
   $bloom = new Bloom::Faster({m=>$m,k=>$hashfct});

   $bloom = &contigsToBloom($assemblyfile,$k,$bloom);
   $date = `date`;
   chomp($date);
   my $writing_tigbloom_message = "\n=>Writing Bloom filter to disk ($bfout): $date\n";
   print $writing_tigbloom_message;
   print LOG $writing_tigbloom_message;
   $assemblyruninfo.=$writing_tigbloom_message;
   $bloom->to_file($bfout);###write BF to disk
 }else{
   my $bffile_message="";
   if(! -e $bf_file){
      $bffile_message = "\nInvalid file: $bf_file -- fatal\n";
      print $bffile_message;
      print LOG $bffile_message;
      exit;
   }else{
      $bffile_message = "Checking Bloom filter file $bf_file...ok\n";
      print $bffile_message;
      print LOG $bffile_message;
   }
   my $bfreusemessage = "A Bloom filter was supplied ($bf_file) and will be used instead of building a new one from -f $assemblyfile\n";
   print LOG $bfreusemessage;
   print $bfreusemessage;
   $bloom = new Bloom::Faster($bf_file);
 }
}###no BF functionality


#-------------------------------------------------
$date = `date`;
chomp($date);

my $reading_reads_message = "\n=>Reading long reads, building hash table: $date\n";
print $reading_reads_message;
print LOG $reading_reads_message;
$assemblyruninfo.=$reading_reads_message;

my $matepair;
my $pairs;
my $ctrd = 0;
my $initpos=$offset;
my $totalpairs = 0;
my $delta = $frag_dist - ( 2 * $k);
my $file_ct = 0;

print "\nReads processed k=$k, dist=$frag_dist, offset=$initpos nt, sliding step=$step nt:\n";
print LOG "\nReads processed k=$k, dist=$frag_dist, offset=$initpos nt, sliding step=$step nt:\n";

open(FOF,$longfile) || die "Can't open file of filenames $longfile for reading -- fatal\n";
while(<FOF>){
   chomp;
   $file_ct++;
   #if(/\.bam/i){
   #   if (! -e $SAMPATH){#only care if there are bam files to read
   #      my $samtool_err_mess = "The executable $SAMPATH does not exist, please revise -- fatal.\n";
   #      print LOG $samtool_err_mess;
   #      print $samtool_err_mess;
   #      close LOG;
   #      exit;
   #   }
   #   ($set,$bin,$sumall,$ctall) = &readBAM($sumall,$ctall,$set,$bin,$_,$encoded,$seedsplit,$r_clip,$q_clip,$c_clip,$e_ascii,$file_ct,$ct_fof_line,$targetwordlen);
   #}else{
   #($set,$bin,$sumall,$ctall) = &readFastaFastq($sumall,$ctall,$set,$bin,$_,$encoded,$seedsplit,$r_clip,$q_clip,$c_clip,$e_ascii,$file_ct,$ct_fof_line,$targetwordlen);
   if(! -e $_){ die "WARNING: Your file $_ does not exist -- fatal.\n";    }
   if($bfoff){
      ($matepair,$pairs,$ctrd) = &readFastaFastqBFoff($file_ct,$ct_fof_line,$ctrd,$_,$frag_dist,$k,$step,$matepair,$delta,$initpos,$bloom,$bfoff);
   }else{
      ($matepair,$pairs,$ctrd) = &readFastaFastq($file_ct,$ct_fof_line,$ctrd,$_,$frag_dist,$k,$step,$matepair,$delta,$initpos,$bloom,$bfoff);
   }
   $totalpairs+=$pairs;
   #}
}
close FOF;

#-------------------------------------------------
$date = `date`;
chomp($date);

my $reading_seqs_message = "\n=>Reading sequence contigs (to scaffold), tracking k-mer positions: $date\n";
print $reading_seqs_message;
print LOG $reading_seqs_message;
$assemblyruninfo.=$reading_seqs_message;
my ($track_all,$tig_length,$tignames) = &readContigs($assemblyfile,$matepair,$k);

#-------------------------------------------------
$date = `date`;
chomp($date);

open (SC, ">$scaffold") || die "\nCan't write to $scaffold -- fatal\n";
my $sc_start_message = "\n=>Scaffolding initiated: $date\n";
print $sc_start_message;
print LOG $sc_start_message;
$assemblyruninfo.= $sc_start_message . "\n";

my $contigpairs  = &pairContigs($longfile,$matepair,$track_all,$tig_length,$issues,$distribution,$totalpairs,$verbose);
&buildScaffolds($contigpairs,$tig_length,$verbose);
 
close SC;
$date = `date`;
chomp($date);

my $sc_end_message = "=>Scaffolding ended: $date\n";
print $sc_end_message;
print LOG $sc_end_message;
$assemblyruninfo.= $sc_end_message . "\n";
print "Scaffolds layout in: $scaffold\n";

#-------------------------------------------------
$date = `date`;
chomp($date);

my $sc_fasta_message = "=>Making FASTA file: $date\n";
print $sc_fasta_message;
print LOG $sc_fasta_message;
$assemblyruninfo.= $sc_fasta_message . "\n";

my $tighash = &readContigsMemory($assemblyfile);
my $scaffold_fasta = &buildScaffoldFasta($scaffold,$tighash);
print "Scaffolds FASTA in: $scaffold_fasta\n";

};###end eval block

#-------------------------------------------------
$date = `date`;
chomp($date);

if($@){
   my $message = $@;
   my $failure = "\nSomething went wrong running $0 $date\n$message\n";
   print $failure;
   print LOG $failure;
   $assemblyruninfo.=$failure . "\n";
}else{
   my $success = "\nScaffolding executed normally $date\n";
   print $success;
   print LOG $success;
   $assemblyruninfo.=$success . "\n";
}
close LOG;

die "$0 $version terminated successfully on $date\n";

# manually free the data structures 
$bloom->DESTROY;

exit 1;


###for dev. test purposes
eval{
   my $wdir = `pwd`;
   chomp($wdir);
   my $smtp = Net::SMTP->new('mailhost');
   $smtp->mail("LINKS\@bcgsc.ca");
   $smtp->to($dev);
   $smtp->data();
   $smtp->datasend("Subject: Your LINKS run\n");
   $smtp->datasend("At: $wdir\n");
   $smtp->datasend($assemblyruninfo);
   $smtp->dataend();
   $smtp->quit;
};

exit;


#----------------
sub contigsToBloom{
   my ($file,$k,$bloom) = @_;

   my $prevhead = "";
   my $seq = "";
   my $cttig=0;
   open(IN,$file) || die "Error reading $file -- fatal.\n";

   print "\nContigs processed k=$k:\n";
   print LOG "\nContigs processed k=$k:\n";
   ###
   while(<IN>){
      if(/^\>(\S+)/){
         my $head=$1;

         if ($head ne $prevhead && $seq ne '' && $prevhead ne ''){
            $cttig++;
            print "\r$cttig";
            $|++;
            $bloom = &kmerizeContigBloom(uc($seq),$bloom,$k,$cttig,0);
            my $revcomp = &reverseComplement(uc($seq));
            $bloom = &kmerizeContigBloom($revcomp,$bloom,$k,$cttig,1);
         }
         $seq = '';
         $prevhead = $head;
      }else{
         $seq .= $_;
      }
   }
   $cttig++;
   print "\r$cttig";
   print LOG "$cttig\n\n";
   $|++;
   $bloom = &kmerizeContigBloom(uc($seq),$bloom,$k,$cttig,0);
   my $revcomp = &reverseComplement(uc($seq));
   $bloom = &kmerizeContigBloom($revcomp,$bloom,$k,$cttig,1);
   ###
   close IN;

   return $bloom;
}


#----------------
sub readContigs{
   my ($file,$matepair,$k) = @_;

   my ($track_all,$tig_length,$tignames);
   my $prevhead = "";
   my $seq = "";
   my $cttig=0;
   open(IN,$file) || die "Error reading $file -- fatal.\n";

   print "\nContigs processed k=$k:\n";
   print LOG "\nContigs processed k=$k:\n";
   ###
   while(<IN>){
      if(/^\>(\S+)/){
         my $head=$1;

         if ($head ne $prevhead && $seq ne '' && $prevhead ne ''){
            $cttig++;
            $tignames->{$cttig}=$prevhead;
            print "\r$cttig";
            $|++;

            $track_all = &kmerizeContig(uc($seq),$track_all,$matepair,$k,$cttig,0);
            my $revcomp = &reverseComplement(uc($seq));
            $track_all = &kmerizeContig($revcomp,$track_all,$matepair,$k,$cttig,1);
            $tig_length->{$cttig} = length($seq);
         }
         $seq = '';
         $prevhead = $head;
      }else{
         $seq .= $_;
      }
   }
   $cttig++;
   $tignames->{$cttig}=$prevhead;
   print "\r$cttig";
   print LOG "$cttig\n\n";
   $|++;
   $track_all = &kmerizeContig(uc($seq),$track_all,$matepair,$k,$cttig,0);
   my $revcomp = &reverseComplement(uc($seq));
   $track_all = &kmerizeContig($revcomp,$track_all,$matepair,$k,$cttig,1);
   $tig_length->{$cttig} = length($seq);
   ###
   close IN;

   return $track_all,$tig_length,$tignames;
}

#----------------
sub readFastaFastq{

   my ($file_ct,$ct_fof_line,$ctrd,$file,$frag_dist,$k,$step,$matepair,$delta,$initpos,$bloom,$bfoff) = @_;

   ###
   my $readinput_message = "\nReads processed from file $file_ct/$ct_fof_line, $file:\n";
   print $readinput_message;
   print LOG $readinput_message;

   open(IN,$file) || die "Error reading $file -- fatal\n";

   my $prevhead = "";
   my $seq = "";
   my $pairs = 0;
   my $quad = 0;

   LINE:
   while(<IN>){
      chomp;
      $quad++;
      if(/^([ACGTMRWSYKVHDBN]+)$/i && $quad<4){$quad=2;}
#      print "QUAD$quad $_ <<<\n";
      #print "$quad\n";
      if($quad==1 || $quad==3){
         #print "1||3 $_\n";
         next LINE if($_ eq "+");
         my $head = $1 if(/^\S{1}(\S+)/);
#print "$quad **$_** $prevhead ne  && $head ne $prevhead\n";
         if($prevhead ne "" && $head ne $prevhead){# && length($seq)>=50){
            #my $qclipflag = 0;
            #if($qual ne ""){$qclipflag = $r_clip;}
            #print "loadSeq: $prev with $seq and $qual\n";
#print ">$prevhead QUAD:$quad\n$seq\n";
            ($matepair,$pairs) = &kmerize(uc($seq),$frag_dist,$k,$matepair,$step,$prevhead,$delta,$initpos,$pairs,$bloom,$bfoff);
            $seq = "";
            $quad=1;
         }
         $ctrd++;
         print "\r$ctrd";
         $|++;
         $prevhead = $head;
      }elsif($quad==2){
         #print "SEQ $_\n";
         $seq .= $_;
      }elsif($quad==4){
         $quad=0;
      }
   }
   if($prevhead ne ""){# && length($seq)>=50){
      ($matepair,$pairs) = &kmerize(uc($seq),$frag_dist,$k,$matepair,$step,$prevhead,$delta,$initpos,$pairs,$bloom,$bfoff);
   }

   close IN;
   print LOG "$ctrd\n\n";

   #####

   print "\nExtracted $pairs total $k-mer pairs\n";  
   print LOG "\nExtracted $pairs total $k-mer pairs\n"; 

   return $matepair,$pairs,$ctrd;
}

#----------------
sub kmerize{

   my ($seq,$frag_dist,$k,$matepair,$step,$head,$delta,$initpos,$pairs,$bloom,$bfoff) = @_;

   for(my $pos=$initpos;$pos<=(length($seq)-$frag_dist);$pos+=$step){

      my $rd1 = substr($seq,$pos,$k);
      my $secondstart = $pos+$k+$delta;
      my $rd2ss = substr($seq,$secondstart,$k);
      my $rd2 = &reverseComplement($rd2ss);
      if(defined $matepair->{$rd2}{$rd1}){my $trd1=$rd2;my $trd2=$rd1;$rd1=$trd1;$rd2=$trd2;}

      if ($bloom->check($rd1) && $bloom->check($rd2)) {###Don't bother hashing k-mer pairs if assembly doesn't have these kmers
         $matepair->{$rd1}{$rd2}{'is'} = $frag_dist;
         #$matepair->{$rd1}{$rd2}{'rd'}{$head}++;  ### this will be used to track uniqueness in pairing (should expect 1 pair per [long]read and one [long]read with that specific pair. HAS LITTLE TO NO EFFECT BUT REQ MORE MEM
         $matepair->{$rd1}{$rd2}{'bt'} = 0;
         $pairs++;
      }
   }
   return $matepair,$pairs;
}

#----------------
sub readFastaFastqBFoff{

   my ($file_ct,$ct_fof_line,$ctrd,$file,$frag_dist,$k,$step,$matepair,$delta,$initpos,$bloom) = @_;

   ###
   my $readinput_message = "\nReads processed from file $file_ct/$ct_fof_line, $file:\n";
   print $readinput_message;
   print LOG $readinput_message;

   open(IN,$file) || die "Error reading $file -- fatal\n";

   my $prevhead = "";
   my $seq = "";
   my $pairs = 0;
   my $quad = 0;
   LINE:
   while(<IN>){
      chomp;
      $quad++;
      if(/^([ACGTMRWSYKVHDBN]+)$/i && $quad<4){$quad=2;}
#      print "QUAD$quad $_ <<<\n";
      #print "$quad\n";
      if($quad==1 || $quad==3){
         #print "1||3 $_\n";
         next LINE if($_ eq "+");
         my $head = $1 if(/^\S{1}(\S+)/);
#print "$quad **$_** $prevhead ne  && $head ne $prevhead\n";
         if($prevhead ne "" && $head ne $prevhead){# && length($seq)>=50){
            #my $qclipflag = 0;
            #if($qual ne ""){$qclipflag = $r_clip;}
            #print "loadSeq: $prev with $seq and $qual\n";
#print ">$prevhead QUAD:$quad\n$seq\n";
            ($matepair,$pairs) = &kmerizeBFoff(uc($seq),$frag_dist,$k,$matepair,$step,$prevhead,$delta,$initpos,$pairs,$bloom);
            $seq = "";
            $quad=1;
         }
         $ctrd++;
         print "\r$ctrd";
         $|++;
         $prevhead = $head;
      }elsif($quad==2){
         #print "SEQ $_\n";
         $seq .= $_;
      }elsif($quad==4){
         $quad=0;
      }
   }
   if($prevhead ne ""){# && length($seq)>=50){
      ($matepair,$pairs) = &kmerizeBFoff(uc($seq),$frag_dist,$k,$matepair,$step,$prevhead,$delta,$initpos,$pairs,$bloom);
   }

   close IN;
   print LOG "$ctrd\n\n";

   #####

   print "\nExtracted $pairs total $k-mer pairs\n";
   print LOG "\nExtracted $pairs total $k-mer pairs\n";

   return $matepair,$pairs,$ctrd;
}

#----------------
sub kmerizeBFoff{

   my ($seq,$frag_dist,$k,$matepair,$step,$head,$delta,$initpos,$pairs,$bloom) = @_;

   for(my $pos=$initpos;$pos<=(length($seq)-$frag_dist);$pos+=$step){

      my $rd1 = substr($seq,$pos,$k);
      my $secondstart = $pos+$k+$delta;
      my $rd2ss = substr($seq,$secondstart,$k);
      my $rd2 = &reverseComplement($rd2ss);
      if(defined $matepair->{$rd2}{$rd1}){my $trd1=$rd2;my $trd2=$rd1;$rd1=$trd1;$rd2=$trd2;}

      $matepair->{$rd1}{$rd2}{'is'} = $frag_dist;
      #$matepair->{$rd1}{$rd2}{'rd'}{$head}++;  ### this will be used to track uniqueness in pairing (should expect 1 pair per [long]read and one [long]read with that specific pair. HAS LITTLE TO NO EFFECT BUT REQ MORE MEM
      $matepair->{$rd1}{$rd2}{'bt'} = 0;
      $pairs++;
   }
   return $matepair,$pairs;
}

#----------------
sub kmerizeContigBloom{
   my ($seq,$bloom,$k,$head,$rc) = @_;

   for(my $pos=0;$pos<=(length($seq)-$k);$pos++){
      my $kmer = substr($seq,$pos,$k);
      $bloom->add($kmer);
   }
   return $bloom;
}

#----------------
sub kmerizeContig{

   my ($seq,$track_all,$matepair,$k,$head,$rc) = @_;

   for(my $pos=0;$pos<=(length($seq)-$k);$pos++){
      my $rd = substr($seq,$pos,$k);
      if(defined $matepair->{$rd}){
         $track_all->{$rd}{'tig'}   = $head;
         $track_all->{$rd}{'start'} = $pos;
         $track_all->{$rd}{'end'}   = $pos + $k;
         if($rc==1){
            $track_all->{$rd}{'start'} = length($seq) - $pos;
            $track_all->{$rd}{'end'}   = length($seq) - ($pos + $k);
         }
         $track_all->{$rd}{'multiple'}++;
      }
   }
   return $track_all;
}

#-----------------------
sub reverseComplement{
   $_ = shift;
   $_ = uc();
   tr/ATGCYRKMBDHV/TACGRYMKVHDB/; 
   return (reverse());
}

#------------------------------------
#Order and orient contigs into scaffolds
sub buildScaffolds{

   my ($pair, $tig_length, $contig_size_cutoff, $verbose) = @_;

   my $seen_it;
   my $sc_ct = 0;
 
   #print SC "Scaffold Number,Scaffold Size (only contig lengths considered),Scaffold Chain: e.g. _f127z7068k12a0.58m42_r3090z62k7r0.14m76_  means: contig127(+ strand=f), size 7068 (z) has 12 links (k), link ratio of 0.58 (a) and with a mean gap/overlap of 42nt (m)  with reverse (r) of contig3090 (size 62) on the right.\n";

   SEED:
   foreach my $tig (sort {$tig_length->{$b}<=>$tig_length->{$a}} keys %$tig_length){

      my $ftig = "f" . $tig;
      my $rtig = "r" . $tig;

      if(! defined $seen_it->{$tig}){##should prevent re-using a contig as seed if it's already been incorporated into a scaffold

         $sc_ct++;

         my $chainleft = "";
          
         my $ori_chainright = $ftig . "Z" . $tig_length->{$tig};
         my $chainright = $ori_chainright;
         my $total = $tig_length->{$tig};

         ($total, $chainright, $seen_it) = &computeLayout("R", $chainright, $ftig, $pair, $tig_length, $total, $seen_it, $tig);
         ($total, $chainleft, $seen_it) = &computeLayout("L", $chainleft, $rtig, $pair, $tig_length, $total, $seen_it, $tig);

         $seen_it->{$tig}++;

         delete $pair->{$ftig};
         delete $pair->{$rtig};
         delete $tig_length->{$tig};

         my $scaffold = $chainleft . $chainright;
         print SC "scaffold" . $sc_ct . ",$total,$scaffold\n" if($total >= $contig_size_cutoff);
      }
   }
}

#------------------------------------
# links contigs together into a chain - must satisfy user-defined criterions (-k -a)
sub computeLayout{

   my ($ext, $chain, $tig, $pair, $tig_length, $total, $seen_it, $orig_tig_number) = @_;

   my $orig_tig = $tig;
   my $extension = 1;

   EXTENSION:
   while($extension){

      my $tnum = $1 if($tig=~/[fr](\d+)/);
      my $tnumf = "f" . $tnum;
      my $tnumr = "r" . $tnum;

      if(! defined $seen_it->{$tnum}){

         $seen_it->{$tnum}++ if($tnumf ne $orig_tig);

         print "Attempt to extend $tig\n" if ($verbose);      
         my $list = $pair->{$tig};
         my ($match1,$link1,$gaps1,$match2,$link2,$gaps2,$cntloop)=("",0,0,"",0,0,0);

         LINK:
         foreach my $match (sort {$list->{$b}{'links'}<=>$list->{$a}{'links'}} keys %$list){

            if($cntloop){
               ($match2,$link2,$gaps2) = ($match,$list->{$match}{'links'},$list->{$match}{'gaps'});
               print "$tig links second best $match2 (links:$link2 total sz:$gaps2)\n" if ($verbose);
               last LINK;
            }else{
               ($match1,$link1,$gaps1) = ($match,$list->{$match}{'links'},$list->{$match}{'gaps'});
               print "$tig links best $match1 (links:$link1 total sz:$gaps1)\n" if ($verbose);
            }
            $cntloop++;
         }

         ###ratio
         my $ratio = 0.00;
         $ratio = $link2 / $link1 if ($link1);        ## relative ratio of the two most abundant contig pairs
         if ($ratio =~ /(\d+\.\d{2})/){$ratio = $1;}
         ###mean
         my $mean = 0;
         $mean = int($gaps1 / $link1) if ($link1);

         my $tempnum = $1 if($match1 =~ /[fr](\d+)/);

         #### Assessment
         if(defined $seen_it->{$tempnum} || $link1 < $min_links || $ratio > $max_link_ratio || $tempnum == $orig_tig_number){
            $extension = 0;
            print "defined seen_it->{ $tempnum } || $link1 < $min_links || $ratio > $max_link_ratio\n L1:$link1 L2:$link2  M1:$match1 M2:$match2 G1:$gaps1 G2:$gaps2 "  if ($verbose);

            last EXTENSION;
         }{### pass filter.. does this contig 
            print "$ext extension.  mean: $mean links:$link1 linkratio:$ratio\n" if ($verbose);

            if($ext eq "R"){
               $chain .= "k" . $link1 . "a" . $ratio . "m" . $mean . "_" . $match1 . "z" . $tig_length->{$tempnum};
            }else{
               my $temp_match = "";
               if($match1 =~ /^r(\d+)/){$temp_match = "f" . $1;}else{$temp_match = "r". $1;}            
               $chain = $temp_match . "z" . $tig_length->{$tempnum} . "k" . $link1 . "a" . $ratio . "m" . $mean . "_" . $chain;
            }   
            $total += $tig_length->{$tempnum};

            print "NEXT TIG TO LOOK AT= $match1\n" if ($verbose);
            $tig = $match1;
            $extension = 1; 
          
            print "Will flag $tnum as seen  (only if $tnumf != $orig_tig)." if ($verbose);
   
            if($tnumf ne $orig_tig){
               delete $pair->{$tnumf};
               delete $pair->{$tnumr};
               delete $tig_length->{$tnum};
            }else{
               delete $pair->{$tnumf};
            }
         }
      }else{
         print "NO MORE MATCH FOR $tig in hash: pair>>\n" if ($verbose);
         $extension = 0;
         last EXTENSION;
      }
   }### pair is defined
   return $total, $chain, $seen_it;
}

#------------------------------------
sub getDistance{

   my ($insert_size, $length_i, $start_i, $start_j) = @_;

   # L  ------  --------- R
   # i    ->        <-    j
   #      ....  ......    insert_span
   #      ============    insert_size

   my $insert_span = ($length_i - $start_i) + $start_j;
   my $gap_or_overlap = $insert_size - $insert_span;

   return $gap_or_overlap;
}

#-----------------
#build contig pairs based on template information  -  must satisfy user-defined criterions (-d -e)
sub pairContigs{

   my ($longfile,$matepair,$track,$tig_length,$issues,$distribution,$totalpairs,$verbose) = @_;
   my ($ct_illogical, $ct_ok_contig, $ct_ok_pairs, $ct_problem_pairs, $ct_iz_issues, $ct_single, $ct_both)= (0,0,0,0,0,0,0);
   my $ct_illogical_hash;
   my $ct_ok_contig_hash;
   my $ct_ok_pairs_hash;
   my $ct_problem_pairs_hash;
   my $ct_iz_issues_hash;
   my $ct_single_hash;
   my $ct_both_hash;

   my ($pair,$err,$track_insert);

   print "Pairing contigs...\n" if ($verbose);

   open(PET, ">$issues") || die "Can't open $issues for writing -- fatal\n";

   foreach my $read_a (keys %$matepair){ 

      my $mateslist = $matepair->{$read_a};

      foreach my $read_b (keys %$mateslist){

          if($matepair->{$read_a}{$read_b}{'bt'}==0 && $track->{$read_a}{'multiple'}==1 && $track->{$read_b}{'multiple'}==1){ ###This has little if no effect, but negative for some odd reason

            ##below indicates this specific pair has been seen
            $matepair->{$read_a}{$read_b}{'bt'}=1;

            my $insert_size = $mateslist->{$read_b}{'is'};
            my $min_allowed = -1 * ($insert_stdev * $insert_size);
            my ($low_iz, $up_iz) = ($insert_size + $min_allowed, $insert_size - $min_allowed);

            print "Pair read1=$read_a read2=$read_b\n" if ($verbose);

            if(defined $track->{$read_a}{'tig'} && defined $track->{$read_b}{'tig'}){### both pairs assembled

               $ct_both++;
               $ct_both_hash->{$insert_size}++;

               my $tig_a = $track->{$read_a}{'tig'};
               my $tig_b = $track->{$read_b}{'tig'};

               my $ftig_a = "f" . $tig_a;
               my $ftig_b = "f" . $tig_b;

               my $rtig_a = "r" . $tig_a;
               my $rtig_b = "r" . $tig_b;

               my $A_length = $tig_length->{$tig_a};
               my $A_start = $track->{$read_a}{'start'};
               my $A_end = $track->{$read_a}{'end'};
 
               my $B_length = $tig_length->{$tig_b};
               my $B_start = $track->{$read_b}{'start'} ;
               my $B_end = $track->{$read_b}{'end'};

               if ($tig_a != $tig_b){####paired reads located on <> contigs

                  ####Determine most likely possibility
                  if ($track->{$read_a}{'start'} < $track->{$read_a}{'end'}){

                     if ($track->{$read_b}{'end'} < $track->{$read_b}{'start'}){####-> <- :::  A-> <-B  /  rB -> <- rA
                         my $d = &getDistance($insert_size, $A_length, $A_start, $B_start);
                         print "A-> <-B  WITH $tig_a -> <- $tig_b GAP $d A=$A_length ($A_start-$A_end) B=$B_length ($B_start-$B_end) Alen, Astart,Bstart\n" if($verbose);
                         if($d >= $min_allowed){
                            $pair->{$ftig_a}{$ftig_b}{'links'}++;
                            $pair->{$ftig_a}{$ftig_b}{'gaps'} += $d;                  
                            $pair->{$rtig_b}{$rtig_a}{'links'}++;
                            $pair->{$rtig_b}{$rtig_a}{'gaps'} += $d;
                            $ct_ok_pairs++;
                            $ct_ok_pairs_hash->{$insert_size}++;
                         }else{
                            my $err_pair = $ftig_a . "-". $ftig_b;
                            $err->{$err_pair}{'links'}++;
                            $err->{$err_pair}{'gaps'} += $d;
                            $ct_problem_pairs++;
                            $ct_problem_pairs_hash->{$insert_size}++;
                            print PET "Pairs unsatisfied in distance within a contig pair.  A-> <-B  WITH tig#$tig_a -> $d <- tig#$tig_b, A=$A_length nt (start:$A_start, end:$A_end) B=$B_length nt (start:$B_start, end:$B_end) CALCULATED DISTANCE APART: $d < $min_allowed\n";
                         }
                      }else{#### -> -> ::: A-> <-rB  / B-> <-rA 
                         my $rB_start = $B_length - $B_start;
                         my $d = &getDistance($insert_size, $A_length, $A_start, $rB_start);
                         print "A-> <-rB  WITH $tig_a -> <- r.$tig_b GAP $d A=$A_length ($A_start-$A_end) B=$B_length ($B_start-$B_end) Alen,Astart,rBstart\n" if($verbose);
                         if($d >= $min_allowed){
                            $pair->{$ftig_a}{$rtig_b}{'links'}++;
                            $pair->{$ftig_a}{$rtig_b}{'gaps'} += $d;
                            $pair->{$ftig_b}{$rtig_a}{'links'}++;
                            $pair->{$ftig_b}{$rtig_a}{'gaps'} += $d;
                            $ct_ok_pairs++;
                            $ct_ok_pairs_hash->{$insert_size}++;
                         }else{
                            my $err_pair = $ftig_a . "-". $rtig_b;
                            $err->{$err_pair}{'links'}++;
                            $err->{$err_pair}{'gaps'} += $d;
                            $ct_problem_pairs++;
                            $ct_problem_pairs_hash->{$insert_size}++;
                            print PET "Pairs unsatisfied in distance within a contig pair.  A-> <-rB  WITH tig#$tig_a -> $d <- tig#r.$tig_b, A=$A_length  nt (start:$A_start, end:$A_end) B=$B_length nt (start:$B_start, end:$B_end) CALCULATED DISTANCE APART: $d < $min_allowed\n";
                         }
                      }
                  }else{

                     if ($track->{$read_b}{'end'} > $track->{$read_b}{'start'}){####<-  -> ::: B-> <-A / rA -> <- rB
                        my $d = &getDistance($insert_size, $B_length, $B_start, $A_start);
                        print "B-> <-A  WITH $tig_b -> <- $tig_a GAP $d A=$A_length ($A_start-$A_end) B=$B_length ($B_start-$B_end) Blen,Bstart,Astart\n" if($verbose);
                        if($d >= $min_allowed){
                           $pair->{$ftig_b}{$ftig_a}{'links'}++;
                           $pair->{$ftig_b}{$ftig_a}{'gaps'} += $d;
                           $pair->{$rtig_a}{$rtig_b}{'links'}++;
                           $pair->{$rtig_a}{$rtig_b}{'gaps'} += $d;
                           $ct_ok_pairs++;
                           $ct_ok_pairs_hash->{$insert_size}++;
                        }else{
                           my $err_pair = $ftig_b . "-". $ftig_a;
                           $err->{$err_pair}{'links'}++;
                           $err->{$err_pair}{'gaps'} += $d;
                           $ct_problem_pairs++;
                           $ct_problem_pairs_hash->{$insert_size}++;
                           print PET "Pairs unsatisfied in distance within a contig pair.  B-> <-A  WITH tig#$tig_b -> $d <- tig#$tig_a, B=$B_length nt (start:$B_start, end:$B_end) A=$A_length nt (start:$A_start, end:$A_end) CALCULATED DISTANCE APART: $d < $min_allowed\n";
                        }
                     }else{                          ####<- <-  :::  rB-> <-A / rA-> <-B
                        my $rB_start = $B_length - $B_start;
                        my $d = &getDistance($insert_size, $B_length, $rB_start, $A_start);
                        print "rB-> <-A WITH r.$tig_b -> <- $tig_a GAP $d A=$A_length ($A_start-$A_end) B=$B_length ($B_start-$B_end) Blen,rBstart,Astart\n" if($verbose);
                        if($d >= $min_allowed){
                           $pair->{$rtig_b}{$ftig_a}{'links'}++;
                           $pair->{$rtig_b}{$ftig_a}{'gaps'} += $d;
                           $pair->{$rtig_a}{$ftig_b}{'links'}++;
                           $pair->{$rtig_a}{$ftig_b}{'gaps'} += $d;
                           $ct_ok_pairs++;
                           $ct_ok_pairs_hash->{$insert_size}++;
                        }else{
                           my $err_pair = $rtig_b . "-". $ftig_a;
                           $err->{$err_pair}{'links'}++;
                           $err->{$err_pair}{'gaps'} += $d;
                           $ct_problem_pairs++;
                           $ct_problem_pairs_hash->{$insert_size}++;
                           print PET "Pairs unsatisfied in distance within a contig pair.  rB-> <-A WITH tig#r.$tig_b -> $d <- tig#$tig_a, B=$B_length nt (start:$B_start, end:$B_end) A=$A_length nt (start:$A_start, end:$A_end) CALCULATED DISTANCE APART: $d < $min_allowed\n";
                        }
                     }
                  }
               }else{###Clone, paired reads located on the same contig -- could be used to investigate misassemblies
           
                  print "Pair ($read_a and $read_b) located on same contig $tig_a ($A_length nt)\n" if ($verbose);
                  my $pet_size = 0;

                  if ($A_start > $B_start && ($B_start < $B_end) && ($A_start > $A_end)){    # B --> <-- A
                     $pet_size = $A_start - $B_start;
                     $track_insert->{$pet_size}++;
                     if($pet_size >= $low_iz && $pet_size <= $up_iz){
                        $ct_ok_contig++;
                        $ct_ok_contig_hash->{$insert_size}++;
                     }else{
                        print PET "Pairs unsatisfied in distance within a contig.  Pair ($read_a - $read_b) on contig $tig_a ($A_length nt) Astart:$A_start Aend:$A_end Bstart:$B_start Bend:$B_end CALCULATED DISTANCE APART: $pet_size\n";
                        $ct_iz_issues++;
                        $ct_iz_issues_hash->{$insert_size}++;
                     }
                  }elsif($B_start > $A_start && ($B_start > $B_end) && ($A_start < $A_end)){ # A --> <-- B
                     $pet_size = $B_start - $A_start;
                     $track_insert->{$pet_size}++;
                     if($pet_size >= $low_iz && $pet_size <= $up_iz){
                        $ct_ok_contig++;
                        $ct_ok_contig_hash->{$insert_size}++;
                     }else{
                        print PET "Pairs unsatisfied in distance within a contig.  Pair ($read_a - $read_b) on contig $tig_a ($A_length nt) Astart:$A_start Aend:$A_end Bstart:$B_start Bend:$B_end CALCULATED DISTANCE APART: $pet_size\n";
                        $ct_iz_issues++;
                        $ct_iz_issues_hash->{$insert_size}++;
                     }
                  }else{
                     $ct_illogical++;
                     $ct_illogical_hash->{$insert_size}++;
                     print PET "Pairs unsatisfied in pairing logic within a contig.  Pair ($read_a - $read_b) on contig $tig_a ($A_length nt) Astart:$A_start Aend:$A_end Bstart:$B_start Bend:$B_end\n";
                  }
               }
            }else{###both pairs assembled
               $ct_single++;
               $ct_single_hash->{$insert_size}++;
            }
         }#if unseen
      }#pairing read b
   }#read a

   ### summary of contig pair issues
   print PET "------------- Putative issues with contig pairing - Summary  ----------------\n";
   foreach my $err_pair (sort {$err->{$b}{'links'}<=>$err->{$a}{'links'}} keys %$err){
      my $mean_iz = 0;
      $mean_iz = $err->{$err_pair}{'gaps'} / $err->{$err_pair}{'links'} if ($err->{$err_pair}{'links'});
      print PET "Pair $err_pair has $err->{$err_pair}{'links'} links and mean distance = $mean_iz\n";
   }
   close PET;
 
   my $satisfied = $ct_ok_pairs + $ct_ok_contig;
   my $unsatisfied = $ct_problem_pairs + $ct_iz_issues + $ct_illogical;
   my $ct_both_reads = $ct_both * 2;

   print LOG "\n===========PAIRED K-MER STATS===========\n";
   print LOG "Total number of pairs extracted from -s $longfile: $totalpairs\n";
   print LOG "At least one sequence/pair missing from contigs: $ct_single\n";
   print LOG "Assembled pairs: $ct_both ($ct_both_reads sequences)\n";
   print LOG "\tSatisfied in distance/logic within contigs (i.e. -> <-, distance on target: $ct_ok_contig\n";
   print LOG "\tUnsatisfied in distance within contigs (i.e. distance out-of-bounds): $ct_iz_issues\n";
   print LOG "\tUnsatisfied pairing logic within contigs (i.e. illogical pairing ->->, <-<- or <-->): $ct_illogical\n";
   print LOG "\t---\n";
   print LOG "\tSatisfied in distance/logic within a given contig pair (pre-scaffold): $ct_ok_pairs\n";
   print LOG "\tUnsatisfied in distance within a given contig pair (i.e. calculated distances out-of-bounds): $ct_problem_pairs\n";
   print LOG "\t---\n";
   print LOG "Total satisfied: $satisfied\tunsatisfied: $unsatisfied\n\nBreakdown by insert sizes:\n";

   foreach my $izopt(sort {$a<=>$b} keys %$ct_both_hash){
      print LOG "--------Reads with $izopt bp inserts--------\n";
      my $maopt = -1 * ($insert_stdev * $izopt);
      my ($low_izopt, $up_izopt) = ($izopt + $maopt, $izopt - $maopt);
      print LOG "MIN:$low_izopt MAX:$up_izopt as defined by $izopt * $insert_stdev\n";
      print LOG "At least one sequence/pair missing: $ct_single_hash->{$izopt}\n";
      print LOG "Assembled pairs: $ct_both_hash->{$izopt}\n";
      print LOG "\tSatisfied in distance/logic within contigs (i.e. -> <-, distance on target: $ct_ok_contig_hash->{$izopt}\n";
      print LOG "\tUnsatisfied in distance within contigs (i.e. distance out-of-bounds): $ct_iz_issues_hash->{$izopt}\n";
      print LOG "\tUnsatisfied pairing logic within contigs (i.e. illogical pairing ->->, <-<- or <-->): $ct_illogical_hash->{$izopt}\n";
      print LOG "\t---\n";
      print LOG "\tSatisfied in distance/logic within a given contig pair (pre-scaffold): $ct_ok_pairs_hash->{$izopt}\n";
      print LOG "\tUnsatisfied in distance within a given contig pair (i.e. calculated distances out-of-bounds): $ct_problem_pairs_hash->{$izopt}\n";
   }
   print LOG "============================================\n";

   open (CSV, ">$distribution") || die "Can't open $distribution for writing -- fatal";

   foreach my $is (sort {$a<=>$b} keys %$track_insert){
      print CSV "$is,$track_insert->{$is}\n";
   }

   close CSV;
   return $pair;
}

#-------------------
sub readContigsMemory{
 
   my $file = shift;

   my $fh;
   my $prev="NA";
   my $seq="";
   my $cttig=0;
   open(FA,$file) || die "Cannot open $file for reading -- fatal.\n";
   while(<FA>){
      chomp;
      if (/\>(\S+)/){
         $cttig++;
         my $head=$cttig;
         #$seq =~ s/[BDEFHIJKLMOPQRSUVWXYZ]/N/g;
         if($prev ne $head && $prev ne "NA"){
             $fh->{$prev} = $seq;
         }
         $prev = $head;
         $seq='';
      }elsif(/^(\S+)$/){
         $seq .= uc($1);
      }
   }
   $fh->{$prev} = $seq;
   close FA;

   return $fh;
}

#-------------------
sub buildScaffoldFasta{

   my ($dotscaffold,$fh) = @_;

   open(IN,$dotscaffold) || die "Cannot open $dotscaffold for reading -- fatal.\n";

   my $scaffold_fasta = $dotscaffold . ".fa";

   open(OUT,">$scaffold_fasta") || die "can't write to $scaffold_fasta -- fatal\n";
   my $tot=0;
   my $ct=0;
   my $sct=0;
   while(<IN>){
      chomp;   
      my $sc="";;
      my @a = split(/\,/);
      my @tig;
   
      if($a[2]=~/\_/){
         @tig = split(/\_/,$a[2]);
      }else{
         push @tig, $a[2];
      }

      $sct++;
      my $tigsum=0;
      print OUT ">$_\n";
      foreach my $t (@tig){
         $ct++;
    
         if($t=~/([fr])(\d+)z(\d+)(\S+)?/i){

            my $orient = $1;
            my $tnum=$2;
            my $head = $orient . $tnum;
            my $search = $tnum;
            my $other = $4;
            $tot+= $3;
            $tigsum +=$3;
   
            my $gap = "NA"; 
            my $gapseq = "";
            $gap = $1 if($other=~/m(\-?\d+)/);
 
            my $seq = $fh->{$search};
            $seq = reverseComplement($seq) if($orient eq "r");

            $gapseq = "N" x $gap if($gap > 0);
            $gapseq = "n" if($gap ne "NA" && $gap <= 0 );
            $seq .= $gapseq;

            print OUT "$seq";         
                
         }#tig regex
   
      }#each tig
      print OUT "\n";
   }

   close IN;
   close OUT;

   return $scaffold_fasta;
}

## We hope this code is useful to you -- Please send comments & suggestions to rwarren at bcgsc.ca
